/****************************************************************************
 * Core Library Version 1.7, August 2004
 * Copyright (c) 1995-2004 Exact Computation Project
 * All rights reserved.
 *
 * This file is part of CGAL (www.cgal.org).
 * You can redistribute it and/or modify it under the terms of the GNU
 * Lesser General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version.
 *
 * Licensees holding a valid commercial license may use this file in
 * accordance with the commercial license agreement provided with the
 * software.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * $URL$
 * $Id$
 ***************************************************************************/

#ifndef _CORE_CONFIG_H_
#define _CORE_CONFIG_H_

// disable debug
//#define CORE_DISABLE_DEBUG

// disable inline functions
//#define CORE_DISABLE_INLINE

// disable memory pool
//#define CORE_DISABLE_MEMPOOL

// debug reference counting
//#define CORE_RC_DEBUG 1

#include <CGAL/auto_link/CORE.h>

#include <CGAL/export/CORE.h>

#endif // _CORE_CONFIG_H_
