// Copyright (c) 2006 Foundation for Research and Technology-Hellas (Greece).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
// 
//
// Author(s)     : Menelaos Karavelas <mkaravel@iacm.forth.gr>

#ifndef CGAL_VORONOI_DIAGRAM_2_BASIC_H
#define CGAL_VORONOI_DIAGRAM_2_BASIC_H 1

#include <CGAL/license/Voronoi_diagram_2.h>


#include <CGAL/basic.h>
#include <cstdlib>

#define CGAL_VORONOI_DIAGRAM_2_INS CGAL::VoronoiDiagram_2::Internal

#endif // CGAL_VORONOI_DIAGRAM_2_BASIC_H
