// Copyright (c) 2003,2004,2005,2006  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
// 
//
// Author(s)     : Menelaos Karavelas <mkaravel@iacm.forth.gr>




#ifndef CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H
#define CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H

#include <CGAL/license/Segment_Delaunay_graph_2.h>


#include <CGAL/Segment_Delaunay_graph_2/Voronoi_vertex_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Compare_x_2.h>
#include <CGAL/Segment_Delaunay_graph_2/Compare_y_2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_parallel_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_same_points_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Are_same_segments_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Orientation_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Oriented_side_of_bisector_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Vertex_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Finite_edge_interior_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Infinite_edge_interior_conflict_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Is_degenerate_edge_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Arrangement_type_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Arrangement_type_non_intersecting_C2.h>
#include <CGAL/Segment_Delaunay_graph_2/Oriented_side_C2.h>

#endif // CGAL_SEGMENT_DELAUNAY_GRAPH_2_PREDICATES_C2_H
