// Copyright (c) 2011 GeoemtryFactory
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 3 of the License,
// or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
//
//
// Author(s)     : Sebastien Loriot <Sebastien.Loriot@geometryfactory.com>
//



#ifndef CGAL_SQRT_EXTENSION_FWD_H
#define CGAL_SQRT_EXTENSION_FWD_H

#include <CGAL/tags.h>

namespace CGAL{

template <class NT_,class ROOT_, class AllowCompareDifferentExtensions = ::CGAL::Tag_false, class FilterPredicates = ::CGAL::Tag_false > 
class Sqrt_extension;
  
} //namespace CGAL

#endif // CGAL_SQRT_EXTENSION_FWD_H
