//
// compiler_config.h is included by CGAL headers to load the needed compiler settings.
//
// DO NOT EDIT compiler_config.h. It is generated by CMake.
//

//#define CGAL_CFG_DENORMALS_COMPILE_BUG 1

#define CGAL_CFG_FPU_ROUNDING_MODE_UNWINDING_VC_BUG 1

//#define CGAL_CFG_IEEE_754_BUG 1

//#define CGAL_CFG_ISTREAM_INT_BUG 1

//#define CGAL_CFG_MATCHING_BUG_5 1

#define CGAL_CFG_MATCHING_BUG_6 1

//#define CGAL_CFG_MATCHING_BUG_7 1

//#define CGAL_CFG_MATCHING_BUG_8 1

#define CGAL_CFG_NESTED_CLASS_FRIEND_DECLARATION_BUG 1

//#define CGAL_CFG_NO_LIMITS 1

//#define CGAL_CFG_NO_NEXTAFTER 1

//#define CGAL_CFG_NO_STL 1

//#define CGAL_CFG_NUMERIC_LIMITS_BUG 1

//#define CGAL_CFG_OUTOFLINE_MEMBER_DEFINITION_BUG 1

//#define CGAL_CFG_TEMPLATE_IN_DEFAULT_PARAMETER_BUG 1

//#define CGAL_CFG_TYPENAME_BEFORE_DEFAULT_ARGUMENT_BUG 1

//#define CGAL_CFG_USING_BASE_MEMBER_BUG_2 1

#define CGAL_USE_GMP 1

#define CGAL_USE_MPFR 1

//#define CGAL_USE_GMPXX 1

//#define CGAL_USE_LEDA 1

//#define CGAL_USE_MPFI 1

//#define CGAL_USE_RS 1

//#define CGAL_USE_NTL 1

#define CGAL_BUILD_SHARED_LIBS 1

#define CGAL_USE_CORE 1

