// Copyright (c) 2006  Utrecht University (The Netherlands),
// ETH Zurich (Switzerland),
// INRIA Sophia-Antipolis (France),
// Max-Planck-Institute Saarbruecken (Germany),
// and Tel-Aviv University (Israel).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 3 of the License,
// or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Author(s)     : -

// This file is automatically created by create_internal_release.
// Do not edit manually.

#ifndef CGAL_VERSION_H
#define CGAL_VERSION_H

#define CGAL_xstr(s) #s
#define CGAL_str(s) CGAL_xstr(s)

#define CGAL_VERSION 4.11
#define CGAL_VERSION_NR 1041101000
#define CGAL_SVN_REVISION 99999
#define CGAL_GIT_HASH a7caa5c433f3c0d53553ce2e5fc7f2f1106d3725
#define CGAL_RELEASE_DATE 20170920
#define CGAL_VERSION_STR CGAL_str(CGAL_VERSION)

#endif
