// Copyright (c) 2004  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
// 
//
// Author(s)     : Laurent RINEAU

#ifndef CGAL_MESH_2_FACE_BADNESS_H
#define CGAL_MESH_2_FACE_BADNESS_H

#include <CGAL/license/Mesh_2.h>


namespace CGAL
{
  namespace Mesh_2
  {
    enum Face_badness { NOT_BAD, BAD, IMPERATIVELY_BAD };
  }
}
#endif // CGAL_MESH_2_FACE_BADNESS_H
