// Copyright (c) 2014
// INRIA Saclay-Ile de France (France)
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 3 of the License,
// or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
//
// Author(s)     : Marc Glisse

#ifndef CGAL_KERNEL_D_CARTESIAN_COMPLETE_H
#define CGAL_KERNEL_D_CARTESIAN_COMPLETE_H

#include <CGAL/NewKernel_d/function_objects_cartesian.h>
#include <CGAL/NewKernel_d/Cartesian_per_dimension.h>
#include <CGAL/NewKernel_d/Types/Segment.h>
#include <CGAL/NewKernel_d/Types/Sphere.h>
#include <CGAL/NewKernel_d/Types/Hyperplane.h>
#include <CGAL/NewKernel_d/Types/Aff_transformation.h>
#include <CGAL/NewKernel_d/Types/Line.h>
#include <CGAL/NewKernel_d/Types/Ray.h>
#include <CGAL/NewKernel_d/Types/Iso_box.h>

#endif // CGAL_KERNEL_D_CARTESIAN_COMPLETE_H
