// Copyright (c) 1997-2007  ETH Zurich (Switzerland).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
// 
//
// Author(s)     : Sven Schoenherr
//                 Bernd Gaertner <gaertner@inf.ethz.ch>
//                 Franz Wessendorp
//                 Kaspar Fischer

#ifndef CGAL_QP_SOLVER_BASIC_H
#define CGAL_QP_SOLVER_BASIC_H

#include <CGAL/license/QP_solver.h>


#include <CGAL/basic.h>
#include <CGAL/Quotient.h>
#include <CGAL/QP_solver/assertions.h>
#include <CGAL/QP_solver/debug.h>
#include <boost/functional.hpp>

#endif // CGAL_QP_SOLVER_BASIC_H

// ===== EOF ==================================================================
