# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_CGAL_Alpha_shape_2')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_CGAL_Alpha_shape_2')
    _CGAL_Alpha_shape_2 = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_CGAL_Alpha_shape_2', [dirname(__file__)])
        except ImportError:
            import _CGAL_Alpha_shape_2
            return _CGAL_Alpha_shape_2
        try:
            _mod = imp.load_module('_CGAL_Alpha_shape_2', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _CGAL_Alpha_shape_2 = swig_import_helper()
    del swig_import_helper
else:
    import _CGAL_Alpha_shape_2
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0

import CGAL.CGAL_Kernel
CGAL_VERSION = _CGAL_Alpha_shape_2.CGAL_VERSION
CGAL_VERSION_NR = _CGAL_Alpha_shape_2.CGAL_VERSION_NR
CGAL_SVN_REVISION = _CGAL_Alpha_shape_2.CGAL_SVN_REVISION
CGAL_RELEASE_DATE = _CGAL_Alpha_shape_2.CGAL_RELEASE_DATE
CGAL_VERSION_STR = _CGAL_Alpha_shape_2.CGAL_VERSION_STR
EXTERIOR = _CGAL_Alpha_shape_2.EXTERIOR
SINGULAR = _CGAL_Alpha_shape_2.SINGULAR
REGULAR = _CGAL_Alpha_shape_2.REGULAR
INTERIOR = _CGAL_Alpha_shape_2.INTERIOR
GENERAL = _CGAL_Alpha_shape_2.GENERAL
REGULARIZED = _CGAL_Alpha_shape_2.REGULARIZED
class Object(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Object, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Object, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Object()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def is_Point_2(self):
        return _CGAL_Alpha_shape_2.Object_is_Point_2(self)

    def get_Point_2(self):
        return _CGAL_Alpha_shape_2.Object_get_Point_2(self)

    def is_Point_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Point_3(self)

    def get_Point_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Point_3(self)

    def is_Triangle_2(self):
        return _CGAL_Alpha_shape_2.Object_is_Triangle_2(self)

    def get_Triangle_2(self):
        return _CGAL_Alpha_shape_2.Object_get_Triangle_2(self)

    def is_Triangle_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Triangle_3(self)

    def get_Triangle_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Triangle_3(self)

    def is_Segment_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Segment_3(self)

    def get_Segment_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Segment_3(self)

    def is_Segment_2(self):
        return _CGAL_Alpha_shape_2.Object_is_Segment_2(self)

    def get_Segment_2(self):
        return _CGAL_Alpha_shape_2.Object_get_Segment_2(self)

    def is_Line_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Line_3(self)

    def get_Line_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Line_3(self)

    def is_Line_2(self):
        return _CGAL_Alpha_shape_2.Object_is_Line_2(self)

    def get_Line_2(self):
        return _CGAL_Alpha_shape_2.Object_get_Line_2(self)

    def is_Plane_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Plane_3(self)

    def get_Plane_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Plane_3(self)

    def is_Ray_2(self):
        return _CGAL_Alpha_shape_2.Object_is_Ray_2(self)

    def get_Ray_2(self):
        return _CGAL_Alpha_shape_2.Object_get_Ray_2(self)

    def is_Ray_3(self):
        return _CGAL_Alpha_shape_2.Object_is_Ray_3(self)

    def get_Ray_3(self):
        return _CGAL_Alpha_shape_2.Object_get_Ray_3(self)

    def empty(self):
        return _CGAL_Alpha_shape_2.Object_empty(self)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Object
    __del__ = lambda self: None
Object_swigregister = _CGAL_Alpha_shape_2.Object_swigregister
Object_swigregister(Object)

class Vertex_range(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Vertex_range, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Vertex_range, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Vertex_range(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_setmethods__["first"] = _CGAL_Alpha_shape_2.Vertex_range_first_set
    __swig_getmethods__["first"] = _CGAL_Alpha_shape_2.Vertex_range_first_get
    if _newclass:
        first = _swig_property(_CGAL_Alpha_shape_2.Vertex_range_first_get, _CGAL_Alpha_shape_2.Vertex_range_first_set)
    __swig_setmethods__["second"] = _CGAL_Alpha_shape_2.Vertex_range_second_set
    __swig_getmethods__["second"] = _CGAL_Alpha_shape_2.Vertex_range_second_get
    if _newclass:
        second = _swig_property(_CGAL_Alpha_shape_2.Vertex_range_second_get, _CGAL_Alpha_shape_2.Vertex_range_second_set)
    def __len__(self):
        return 2
    def __repr__(self):
        return str((self.first, self.second))
    def __getitem__(self, index): 
        if not (index % 2):
            return self.first
        else:
            return self.second
    def __setitem__(self, index, val):
        if not (index % 2):
            self.first = val
        else:
            self.second = val
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Vertex_range
    __del__ = lambda self: None
Vertex_range_swigregister = _CGAL_Alpha_shape_2.Vertex_range_swigregister
Vertex_range_swigregister(Vertex_range)

class Face_Interval_3(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Face_Interval_3, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Face_Interval_3, name)
    __repr__ = _swig_repr
    __swig_setmethods__["first"] = _CGAL_Alpha_shape_2.Face_Interval_3_first_set
    __swig_getmethods__["first"] = _CGAL_Alpha_shape_2.Face_Interval_3_first_get
    if _newclass:
        first = _swig_property(_CGAL_Alpha_shape_2.Face_Interval_3_first_get, _CGAL_Alpha_shape_2.Face_Interval_3_first_set)
    __swig_setmethods__["second"] = _CGAL_Alpha_shape_2.Face_Interval_3_second_set
    __swig_getmethods__["second"] = _CGAL_Alpha_shape_2.Face_Interval_3_second_get
    if _newclass:
        second = _swig_property(_CGAL_Alpha_shape_2.Face_Interval_3_second_get, _CGAL_Alpha_shape_2.Face_Interval_3_second_set)
    __swig_setmethods__["third"] = _CGAL_Alpha_shape_2.Face_Interval_3_third_set
    __swig_getmethods__["third"] = _CGAL_Alpha_shape_2.Face_Interval_3_third_get
    if _newclass:
        third = _swig_property(_CGAL_Alpha_shape_2.Face_Interval_3_third_get, _CGAL_Alpha_shape_2.Face_Interval_3_third_set)

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Face_Interval_3(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Face_Interval_3_deepcopy(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Face_Interval_3
    __del__ = lambda self: None
Face_Interval_3_swigregister = _CGAL_Alpha_shape_2.Face_Interval_3_swigregister
Face_Interval_3_swigregister(Face_Interval_3)

class Alpha_shape_2_Face_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Face_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Face_handle, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Face_handle()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def dimension(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_dimension(self)

    def vertex(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_vertex(self, *args)

    def has_vertex(self, c):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_has_vertex(self, c)

    def neighbor(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_neighbor(self, *args)

    def has_neighbor(self, c):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_has_neighbor(self, c)

    def index(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_index(self, *args)

    def set_vertex(self, c1, c2):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_vertex(self, c1, c2)

    def set_vertices(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_vertices(self, *args)

    def set_neighbor(self, c1, c2):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_neighbor(self, c1, c2)

    def set_neighbors(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_neighbors(self, *args)

    def reorient(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_reorient(self)

    def ccw_permute(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_ccw_permute(self)

    def cw_permute(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_cw_permute(self)

    def is_valid(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_is_valid(self, *args)

    def __lt__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___lt__(self, p)

    def __gt__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___gt__(self, p)

    def __le__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___le__(self, p)

    def __ge__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___ge__(self, p)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___ne__(self, p)

    def __hash__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle___hash__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_deepcopy(self, *args)

    def get_alpha(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_get_alpha(self)

    def get_ranges(self, i):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_get_ranges(self, i)

    def set_ranges(self, i, r):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_ranges(self, i, r)

    def set_alpha(self, d):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_set_alpha(self, d)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Face_handle
    __del__ = lambda self: None
Alpha_shape_2_Face_handle_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Face_handle_swigregister
Alpha_shape_2_Face_handle_swigregister(Alpha_shape_2_Face_handle)

class Alpha_shape_2_Vertex_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Vertex_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Vertex_handle, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Vertex_handle()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def point(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_point(self, *args)

    def set_point(self, c):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_set_point(self, c)

    def __lt__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___lt__(self, p)

    def __gt__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___gt__(self, p)

    def __le__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___le__(self, p)

    def __ge__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___ge__(self, p)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___ne__(self, p)

    def __hash__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle___hash__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_deepcopy(self, *args)

    def get_range(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_get_range(self)

    def set_range(self, range):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_set_range(self, range)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Vertex_handle
    __del__ = lambda self: None
Alpha_shape_2_Vertex_handle_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_handle_swigregister
Alpha_shape_2_Vertex_handle_swigregister(Alpha_shape_2_Vertex_handle)

class Alpha_shape_2_Edge(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Edge, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Edge, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Edge(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_setmethods__["first"] = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_first_set
    __swig_getmethods__["first"] = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_first_get
    if _newclass:
        first = _swig_property(_CGAL_Alpha_shape_2.Alpha_shape_2_Edge_first_get, _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_first_set)
    __swig_setmethods__["second"] = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_second_set
    __swig_getmethods__["second"] = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_second_get
    if _newclass:
        second = _swig_property(_CGAL_Alpha_shape_2.Alpha_shape_2_Edge_second_get, _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_second_set)
    def __len__(self):
        return 2
    def __repr__(self):
        return str((self.first, self.second))
    def __getitem__(self, index): 
        if not (index % 2):
            return self.first
        else:
            return self.second
    def __setitem__(self, index, val):
        if not (index % 2):
            self.first = val
        else:
            self.second = val
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Edge
    __del__ = lambda self: None
Alpha_shape_2_Edge_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_swigregister
Alpha_shape_2_Edge_swigregister(Alpha_shape_2_Edge)

class Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def clear(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_clear(self)

    def dimension(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_dimension(self)

    def number_of_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_number_of_vertices(self)

    def number_of_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_number_of_faces(self)

    def infinite_face(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_infinite_face(self, *args)

    def infinite_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_infinite_vertex(self, *args)

    def finite_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_finite_vertex(self, *args)

    def is_infinite(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_is_infinite(self, *args)

    def is_edge(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_is_edge(self, *args)

    def includes_edge(self, va, vb, vbr, fr, i):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_includes_edge(self, va, vb, vbr, fr, i)

    def is_face(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_is_face(self, *args)

    def oriented_side(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_oriented_side(self, c1, c2)

    def side_of_oriented_circle(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_side_of_oriented_circle(self, c1, c2)

    def locate(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_locate(self, *args)

    def finite_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_finite_vertices(self)

    def finite_edges(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_finite_edges(self)

    def finite_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_finite_faces(self)

    def all_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_all_vertices(self)

    def all_edges(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_all_edges(self)

    def all_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_all_faces(self)

    def points(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_points(self)

    def line_walk(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_line_walk(self, *args)

    def incident_faces(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_incident_faces(self, *args)

    def incident_edges(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_incident_edges(self, *args)

    def incident_vertices(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_incident_vertices(self, *args)

    def mirror_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_mirror_vertex(self, *args)

    def mirror_index(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_mirror_index(self, c1, c2)

    def ccw(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_ccw(self, c)

    def cw(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_cw(self, c)

    def circumcenter(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_circumcenter(self, *args)

    def set_infinite_vertex(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_set_infinite_vertex(self, c)

    def is_valid(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_is_valid(self, *args)

    def triangle(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_triangle(self, *args)

    def segment(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_segment(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2
    __del__ = lambda self: None
Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_swigregister
Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2_swigregister(Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2)

class Alpha_shape_2_All_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_All_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_All_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_All_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_All_vertices_iterator
    __del__ = lambda self: None
Alpha_shape_2_All_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_All_vertices_iterator_swigregister
Alpha_shape_2_All_vertices_iterator_swigregister(Alpha_shape_2_All_vertices_iterator)

class Alpha_shape_2_Finite_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Finite_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Finite_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Finite_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Finite_vertices_iterator
    __del__ = lambda self: None
Alpha_shape_2_Finite_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_vertices_iterator_swigregister
Alpha_shape_2_Finite_vertices_iterator_swigregister(Alpha_shape_2_Finite_vertices_iterator)

class Alpha_shape_2_All_faces_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_All_faces_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_All_faces_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_All_faces_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_All_faces_iterator
    __del__ = lambda self: None
Alpha_shape_2_All_faces_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_All_faces_iterator_swigregister
Alpha_shape_2_All_faces_iterator_swigregister(Alpha_shape_2_All_faces_iterator)

class Alpha_shape_2_Finite_faces_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Finite_faces_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Finite_faces_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Finite_faces_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Finite_faces_iterator
    __del__ = lambda self: None
Alpha_shape_2_Finite_faces_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_faces_iterator_swigregister
Alpha_shape_2_Finite_faces_iterator_swigregister(Alpha_shape_2_Finite_faces_iterator)

class Alpha_shape_2_All_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_All_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_All_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_All_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_All_edges_iterator
    __del__ = lambda self: None
Alpha_shape_2_All_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_All_edges_iterator_swigregister
Alpha_shape_2_All_edges_iterator_swigregister(Alpha_shape_2_All_edges_iterator)

class Alpha_shape_2_Finite_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Finite_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Finite_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Finite_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Finite_edges_iterator
    __del__ = lambda self: None
Alpha_shape_2_Finite_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Finite_edges_iterator_swigregister
Alpha_shape_2_Finite_edges_iterator_swigregister(Alpha_shape_2_Finite_edges_iterator)

class Alpha_shape_2_Point_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Point_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Point_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Point_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Point_iterator
    __del__ = lambda self: None
Alpha_shape_2_Point_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Point_iterator_swigregister
Alpha_shape_2_Point_iterator_swigregister(Alpha_shape_2_Point_iterator)

class Alpha_shape_2_Line_face_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Line_face_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Line_face_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Line_face_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Line_face_circulator
    __del__ = lambda self: None
Alpha_shape_2_Line_face_circulator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Line_face_circulator_swigregister
Alpha_shape_2_Line_face_circulator_swigregister(Alpha_shape_2_Line_face_circulator)

class Alpha_shape_2_Face_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Face_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Face_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Face_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Face_circulator
    __del__ = lambda self: None
Alpha_shape_2_Face_circulator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Face_circulator_swigregister
Alpha_shape_2_Face_circulator_swigregister(Alpha_shape_2_Face_circulator)

class Alpha_shape_2_Edge_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Edge_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Edge_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Edge_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Edge_circulator
    __del__ = lambda self: None
Alpha_shape_2_Edge_circulator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Edge_circulator_swigregister
Alpha_shape_2_Edge_circulator_swigregister(Alpha_shape_2_Edge_circulator)

class Alpha_shape_2_Vertex_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Vertex_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Vertex_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Vertex_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Vertex_circulator
    __del__ = lambda self: None
Alpha_shape_2_Vertex_circulator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Vertex_circulator_swigregister
Alpha_shape_2_Vertex_circulator_swigregister(Alpha_shape_2_Vertex_circulator)

class Ref_Alpha_shape_2_Face_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Ref_Alpha_shape_2_Face_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Ref_Alpha_shape_2_Face_handle, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Ref_Alpha_shape_2_Face_handle(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set(self, t):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Face_handle_set(self, t)

    def object(self):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Face_handle_object(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Face_handle_deepcopy(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Ref_Alpha_shape_2_Face_handle
    __del__ = lambda self: None
Ref_Alpha_shape_2_Face_handle_swigregister = _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Face_handle_swigregister
Ref_Alpha_shape_2_Face_handle_swigregister(Ref_Alpha_shape_2_Face_handle)

class Ref_Alpha_shape_2_Vertex_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Ref_Alpha_shape_2_Vertex_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Ref_Alpha_shape_2_Vertex_handle, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Ref_Alpha_shape_2_Vertex_handle(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set(self, t):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Vertex_handle_set(self, t)

    def object(self):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Vertex_handle_object(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Vertex_handle_deepcopy(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Ref_Alpha_shape_2_Vertex_handle
    __del__ = lambda self: None
Ref_Alpha_shape_2_Vertex_handle_swigregister = _CGAL_Alpha_shape_2.Ref_Alpha_shape_2_Vertex_handle_swigregister
Ref_Alpha_shape_2_Vertex_handle_swigregister(Ref_Alpha_shape_2_Vertex_handle)

class Internal_Delaunay_Alpha_shape_2(Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2):
    __swig_setmethods__ = {}
    for _s in [Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, Internal_Delaunay_Alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    for _s in [Internal_Triangulation_2_Internal_Delaunay_Alpha_shape_2]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, Internal_Delaunay_Alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Internal_Delaunay_Alpha_shape_2()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def nearest_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_nearest_vertex(self, *args)

    def get_conflicts_and_boundary(self, p, fout, eout, start):
        return _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_get_conflicts_and_boundary(self, p, fout, eout, start)

    def get_conflicts(self, p, fout, start):
        return _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_get_conflicts(self, p, fout, start)

    def get_boundary_of_conflicts(self, p, out, start):
        return _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_get_boundary_of_conflicts(self, p, out, start)

    def dual(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_dual(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Internal_Delaunay_Alpha_shape_2
    __del__ = lambda self: None
Internal_Delaunay_Alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Internal_Delaunay_Alpha_shape_2_swigregister
Internal_Delaunay_Alpha_shape_2_swigregister(Internal_Delaunay_Alpha_shape_2)

class Alpha_shape_2(Internal_Delaunay_Alpha_shape_2):
    __swig_setmethods__ = {}
    for _s in [Internal_Delaunay_Alpha_shape_2]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    for _s in [Internal_Delaunay_Alpha_shape_2]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set_mode(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_set_mode(self, *args)

    def get_mode(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_get_mode(self)

    def clear(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_clear(self)

    def set_alpha(self, c):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_set_alpha(self, c)

    def get_alpha(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_get_alpha(self)

    def get_nth_alpha(self, c):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_get_nth_alpha(self, c)

    def number_of_alphas(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_number_of_alphas(self)

    def make_alpha_shape(self, range):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_make_alpha_shape(self, range)

    def alpha(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha(self)

    def alpha_find(self, a):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha_find(self, a)

    def alpha_lower_bound(self, a):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha_lower_bound(self, a)

    def alpha_upper_bound(self, a):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha_upper_bound(self, a)

    def number_of_solid_components(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_number_of_solid_components(self, *args)

    def find_optimal_alpha(self, i):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_find_optimal_alpha(self, i)

    def alpha_shape_vertices(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha_shape_vertices(self)

    def alpha_shape_edges(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_alpha_shape_edges(self)

    def classify(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_classify(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2
    __del__ = lambda self: None
Alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_swigregister
Alpha_shape_2_swigregister(Alpha_shape_2)

class Alpha_shape_2_Alpha_shape_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Alpha_shape_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Alpha_shape_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Alpha_shape_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Alpha_shape_vertices_iterator
    __del__ = lambda self: None
Alpha_shape_2_Alpha_shape_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_vertices_iterator_swigregister
Alpha_shape_2_Alpha_shape_vertices_iterator_swigregister(Alpha_shape_2_Alpha_shape_vertices_iterator)

class Alpha_shape_2_Alpha_shape_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Alpha_shape_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Alpha_shape_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Alpha_shape_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Alpha_shape_edges_iterator
    __del__ = lambda self: None
Alpha_shape_2_Alpha_shape_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_shape_edges_iterator_swigregister
Alpha_shape_2_Alpha_shape_edges_iterator_swigregister(Alpha_shape_2_Alpha_shape_edges_iterator)

class Alpha_shape_2_Alpha_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Alpha_shape_2_Alpha_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Alpha_shape_2_Alpha_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Alpha_shape_2_Alpha_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator_next(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator___next__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Alpha_shape_2_Alpha_iterator
    __del__ = lambda self: None
Alpha_shape_2_Alpha_iterator_swigregister = _CGAL_Alpha_shape_2.Alpha_shape_2_Alpha_iterator_swigregister
Alpha_shape_2_Alpha_iterator_swigregister(Alpha_shape_2_Alpha_iterator)

class Weighted_alpha_shape_2_Face_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Face_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Face_handle, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Face_handle()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def dimension(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_dimension(self)

    def vertex(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_vertex(self, *args)

    def has_vertex(self, c):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_has_vertex(self, c)

    def neighbor(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_neighbor(self, *args)

    def has_neighbor(self, c):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_has_neighbor(self, c)

    def index(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_index(self, *args)

    def set_vertex(self, c1, c2):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_vertex(self, c1, c2)

    def set_vertices(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_vertices(self, *args)

    def set_neighbor(self, c1, c2):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_neighbor(self, c1, c2)

    def set_neighbors(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_neighbors(self, *args)

    def reorient(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_reorient(self)

    def ccw_permute(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_ccw_permute(self)

    def cw_permute(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_cw_permute(self)

    def is_valid(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_is_valid(self, *args)

    def __lt__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___lt__(self, p)

    def __gt__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___gt__(self, p)

    def __le__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___le__(self, p)

    def __ge__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___ge__(self, p)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___ne__(self, p)

    def __hash__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle___hash__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_deepcopy(self, *args)

    def get_alpha(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_get_alpha(self)

    def get_ranges(self, i):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_get_ranges(self, i)

    def set_ranges(self, i, r):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_ranges(self, i, r)

    def set_alpha(self, d):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_set_alpha(self, d)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Face_handle
    __del__ = lambda self: None
Weighted_alpha_shape_2_Face_handle_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_handle_swigregister
Weighted_alpha_shape_2_Face_handle_swigregister(Weighted_alpha_shape_2_Face_handle)

class Weighted_alpha_shape_2_Vertex_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Vertex_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Vertex_handle, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Vertex_handle()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def point(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_point(self, *args)

    def set_point(self, c):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_set_point(self, c)

    def __lt__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___lt__(self, p)

    def __gt__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___gt__(self, p)

    def __le__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___le__(self, p)

    def __ge__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___ge__(self, p)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___ne__(self, p)

    def __hash__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle___hash__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_deepcopy(self, *args)

    def get_range(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_get_range(self)

    def set_range(self, range):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_set_range(self, range)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Vertex_handle
    __del__ = lambda self: None
Weighted_alpha_shape_2_Vertex_handle_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_handle_swigregister
Weighted_alpha_shape_2_Vertex_handle_swigregister(Weighted_alpha_shape_2_Vertex_handle)

class Weighted_alpha_shape_2_Edge(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Edge, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Edge, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Edge(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_setmethods__["first"] = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_first_set
    __swig_getmethods__["first"] = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_first_get
    if _newclass:
        first = _swig_property(_CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_first_get, _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_first_set)
    __swig_setmethods__["second"] = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_second_set
    __swig_getmethods__["second"] = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_second_get
    if _newclass:
        second = _swig_property(_CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_second_get, _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_second_set)
    def __len__(self):
        return 2
    def __repr__(self):
        return str((self.first, self.second))
    def __getitem__(self, index): 
        if not (index % 2):
            return self.first
        else:
            return self.second
    def __setitem__(self, index, val):
        if not (index % 2):
            self.first = val
        else:
            self.second = val
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Edge
    __del__ = lambda self: None
Weighted_alpha_shape_2_Edge_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_swigregister
Weighted_alpha_shape_2_Edge_swigregister(Weighted_alpha_shape_2_Edge)

class Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def clear(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_clear(self)

    def dimension(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_dimension(self)

    def number_of_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_number_of_vertices(self)

    def number_of_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_number_of_faces(self)

    def infinite_face(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_infinite_face(self, *args)

    def infinite_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_infinite_vertex(self, *args)

    def finite_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_finite_vertex(self, *args)

    def is_infinite(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_is_infinite(self, *args)

    def is_edge(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_is_edge(self, *args)

    def includes_edge(self, va, vb, vbr, fr, i):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_includes_edge(self, va, vb, vbr, fr, i)

    def is_face(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_is_face(self, *args)

    def oriented_side(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_oriented_side(self, c1, c2)

    def side_of_oriented_circle(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_side_of_oriented_circle(self, c1, c2)

    def locate(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_locate(self, *args)

    def finite_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_finite_vertices(self)

    def finite_edges(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_finite_edges(self)

    def finite_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_finite_faces(self)

    def all_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_all_vertices(self)

    def all_edges(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_all_edges(self)

    def all_faces(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_all_faces(self)

    def points(self):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_points(self)

    def line_walk(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_line_walk(self, *args)

    def incident_faces(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_incident_faces(self, *args)

    def incident_edges(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_incident_edges(self, *args)

    def incident_vertices(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_incident_vertices(self, *args)

    def mirror_vertex(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_mirror_vertex(self, *args)

    def mirror_index(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_mirror_index(self, c1, c2)

    def ccw(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_ccw(self, c)

    def cw(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_cw(self, c)

    def circumcenter(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_circumcenter(self, *args)

    def set_infinite_vertex(self, c):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_set_infinite_vertex(self, c)

    def is_valid(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_is_valid(self, *args)

    def triangle(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_triangle(self, *args)

    def segment(self, *args):
        return _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_segment(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2
    __del__ = lambda self: None
Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_swigregister
Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2_swigregister(Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2)

class Weighted_alpha_shape_2_All_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_All_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_All_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_All_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_All_vertices_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_All_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_vertices_iterator_swigregister
Weighted_alpha_shape_2_All_vertices_iterator_swigregister(Weighted_alpha_shape_2_All_vertices_iterator)

class Weighted_alpha_shape_2_Finite_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Finite_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Finite_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Finite_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Finite_vertices_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Finite_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_vertices_iterator_swigregister
Weighted_alpha_shape_2_Finite_vertices_iterator_swigregister(Weighted_alpha_shape_2_Finite_vertices_iterator)

class Weighted_alpha_shape_2_All_faces_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_All_faces_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_All_faces_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_All_faces_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_All_faces_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_All_faces_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_faces_iterator_swigregister
Weighted_alpha_shape_2_All_faces_iterator_swigregister(Weighted_alpha_shape_2_All_faces_iterator)

class Weighted_alpha_shape_2_Finite_faces_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Finite_faces_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Finite_faces_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Finite_faces_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Finite_faces_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Finite_faces_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_faces_iterator_swigregister
Weighted_alpha_shape_2_Finite_faces_iterator_swigregister(Weighted_alpha_shape_2_Finite_faces_iterator)

class Weighted_alpha_shape_2_All_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_All_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_All_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_All_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_All_edges_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_All_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_All_edges_iterator_swigregister
Weighted_alpha_shape_2_All_edges_iterator_swigregister(Weighted_alpha_shape_2_All_edges_iterator)

class Weighted_alpha_shape_2_Finite_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Finite_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Finite_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Finite_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Finite_edges_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Finite_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Finite_edges_iterator_swigregister
Weighted_alpha_shape_2_Finite_edges_iterator_swigregister(Weighted_alpha_shape_2_Finite_edges_iterator)

class Weighted_alpha_shape_2_Point_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Point_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Point_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Point_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Point_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Point_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Point_iterator_swigregister
Weighted_alpha_shape_2_Point_iterator_swigregister(Weighted_alpha_shape_2_Point_iterator)

class Weighted_alpha_shape_2_Line_face_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Line_face_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Line_face_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Line_face_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Line_face_circulator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Line_face_circulator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Line_face_circulator_swigregister
Weighted_alpha_shape_2_Line_face_circulator_swigregister(Weighted_alpha_shape_2_Line_face_circulator)

class Weighted_alpha_shape_2_Face_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Face_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Face_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Face_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Face_circulator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Face_circulator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Face_circulator_swigregister
Weighted_alpha_shape_2_Face_circulator_swigregister(Weighted_alpha_shape_2_Face_circulator)

class Weighted_alpha_shape_2_Edge_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Edge_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Edge_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Edge_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Edge_circulator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Edge_circulator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Edge_circulator_swigregister
Weighted_alpha_shape_2_Edge_circulator_swigregister(Weighted_alpha_shape_2_Edge_circulator)

class Weighted_alpha_shape_2_Vertex_circulator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Vertex_circulator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Vertex_circulator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Vertex_circulator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator_next(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator_deepcopy(self, *args)

    def prev(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator_prev(self)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Vertex_circulator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Vertex_circulator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Vertex_circulator_swigregister
Weighted_alpha_shape_2_Vertex_circulator_swigregister(Weighted_alpha_shape_2_Vertex_circulator)

class Ref_Weighted_alpha_shape_2_Face_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Ref_Weighted_alpha_shape_2_Face_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Ref_Weighted_alpha_shape_2_Face_handle, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Ref_Weighted_alpha_shape_2_Face_handle(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set(self, t):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Face_handle_set(self, t)

    def object(self):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Face_handle_object(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Face_handle_deepcopy(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Ref_Weighted_alpha_shape_2_Face_handle
    __del__ = lambda self: None
Ref_Weighted_alpha_shape_2_Face_handle_swigregister = _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Face_handle_swigregister
Ref_Weighted_alpha_shape_2_Face_handle_swigregister(Ref_Weighted_alpha_shape_2_Face_handle)

class Ref_Weighted_alpha_shape_2_Vertex_handle(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Ref_Weighted_alpha_shape_2_Vertex_handle, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Ref_Weighted_alpha_shape_2_Vertex_handle, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Ref_Weighted_alpha_shape_2_Vertex_handle(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set(self, t):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Vertex_handle_set(self, t)

    def object(self):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Vertex_handle_object(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Vertex_handle_deepcopy(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Ref_Weighted_alpha_shape_2_Vertex_handle
    __del__ = lambda self: None
Ref_Weighted_alpha_shape_2_Vertex_handle_swigregister = _CGAL_Alpha_shape_2.Ref_Weighted_alpha_shape_2_Vertex_handle_swigregister
Ref_Weighted_alpha_shape_2_Vertex_handle_swigregister(Ref_Weighted_alpha_shape_2_Vertex_handle)

class Weighted_alpha_shape_2_Hidden_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Hidden_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Hidden_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Hidden_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Hidden_vertices_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Hidden_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Hidden_vertices_iterator_swigregister
Weighted_alpha_shape_2_Hidden_vertices_iterator_swigregister(Weighted_alpha_shape_2_Hidden_vertices_iterator)

class Internal_regular_Weighted_alpha_shape_2(Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2):
    __swig_setmethods__ = {}
    for _s in [Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, Internal_regular_Weighted_alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    for _s in [Internal_Triangulation_2_Internal_regular_Weighted_alpha_shape_2]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, Internal_regular_Weighted_alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Internal_regular_Weighted_alpha_shape_2()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def number_of_hidden_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_number_of_hidden_vertices(self)

    def hidden_vertices(self):
        return _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_hidden_vertices(self)

    def weighted_circumcenter(self, *args):
        return _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_weighted_circumcenter(self, *args)

    def dual(self, *args):
        return _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_dual(self, *args)

    def power_test(self, c1, c2):
        return _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_power_test(self, c1, c2)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Internal_regular_Weighted_alpha_shape_2
    __del__ = lambda self: None
Internal_regular_Weighted_alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Internal_regular_Weighted_alpha_shape_2_swigregister
Internal_regular_Weighted_alpha_shape_2_swigregister(Internal_regular_Weighted_alpha_shape_2)

class Weighted_alpha_shape_2(Internal_regular_Weighted_alpha_shape_2):
    __swig_setmethods__ = {}
    for _s in [Internal_regular_Weighted_alpha_shape_2]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2, name, value)
    __swig_getmethods__ = {}
    for _s in [Internal_regular_Weighted_alpha_shape_2]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def set_mode(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_set_mode(self, *args)

    def get_mode(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_get_mode(self)

    def clear(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_clear(self)

    def set_alpha(self, c):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_set_alpha(self, c)

    def get_alpha(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_get_alpha(self)

    def get_nth_alpha(self, c):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_get_nth_alpha(self, c)

    def number_of_alphas(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_number_of_alphas(self)

    def make_alpha_shape(self, range):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_make_alpha_shape(self, range)

    def alpha(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha(self)

    def alpha_find(self, a):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha_find(self, a)

    def alpha_lower_bound(self, a):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha_lower_bound(self, a)

    def alpha_upper_bound(self, a):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha_upper_bound(self, a)

    def number_of_solid_components(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_number_of_solid_components(self, *args)

    def find_optimal_alpha(self, i):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_find_optimal_alpha(self, i)

    def alpha_shape_vertices(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha_shape_vertices(self)

    def alpha_shape_edges(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_alpha_shape_edges(self)

    def classify(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_classify(self, *args)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2
    __del__ = lambda self: None
Weighted_alpha_shape_2_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_swigregister
Weighted_alpha_shape_2_swigregister(Weighted_alpha_shape_2)

class Weighted_alpha_shape_2_Alpha_shape_vertices_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Alpha_shape_vertices_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Alpha_shape_vertices_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Alpha_shape_vertices_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Alpha_shape_vertices_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_swigregister
Weighted_alpha_shape_2_Alpha_shape_vertices_iterator_swigregister(Weighted_alpha_shape_2_Alpha_shape_vertices_iterator)

class Weighted_alpha_shape_2_Alpha_shape_edges_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Alpha_shape_edges_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Alpha_shape_edges_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Alpha_shape_edges_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator___iter__(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator___next__(self)

    def next(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator_next(self, *args)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Alpha_shape_edges_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Alpha_shape_edges_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_shape_edges_iterator_swigregister
Weighted_alpha_shape_2_Alpha_shape_edges_iterator_swigregister(Weighted_alpha_shape_2_Alpha_shape_edges_iterator)

class Weighted_alpha_shape_2_Alpha_iterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Weighted_alpha_shape_2_Alpha_iterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Weighted_alpha_shape_2_Alpha_iterator, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _CGAL_Alpha_shape_2.new_Weighted_alpha_shape_2_Alpha_iterator()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def __iter__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator___iter__(self)

    def next(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator_next(self)

    def __next__(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator___next__(self)

    def deepcopy(self, *args):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator_deepcopy(self, *args)

    def hasNext(self):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator_hasNext(self)

    def __eq__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator___eq__(self, p)

    def __ne__(self, p):
        return _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator___ne__(self, p)
    __swig_destroy__ = _CGAL_Alpha_shape_2.delete_Weighted_alpha_shape_2_Alpha_iterator
    __del__ = lambda self: None
Weighted_alpha_shape_2_Alpha_iterator_swigregister = _CGAL_Alpha_shape_2.Weighted_alpha_shape_2_Alpha_iterator_swigregister
Weighted_alpha_shape_2_Alpha_iterator_swigregister(Weighted_alpha_shape_2_Alpha_iterator)

# This file is compatible with both classic and new-style classes.


