// Copyright (c) 2016  GeometryFactory Sarl (France)
//
// This file is part of CGAL (www.cgal.org); you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 3 of the License,
// or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL$
// $Id$
// 
//
// Author     : Andreas Fabri

#ifndef CGAL_LICENSE_CHECK_H
#define CGAL_LICENSE_CHECK_H

//#define CGAL_LICENSE_WARNING  1   // in order to get a warning during compilation
//#define CGAL_LICENSE_ERROR    1   // in order to get an error during compilation

// if no such define exists, the package is used under the GPL or LGPL
// Otherwise the number encodes the date year/month/day
// Any release made b......
// e.g.
//#define CGAL_AABB_TREE_COMMERCIAL_LICENSE  20140101

#endif // CGAL_LICENSE_CHECK_H
